unit uWipeSpace;

//WipeSpace is serves 2 purposes: first it is a program for wiping the space of
//a reused and freshly installed computers. Second it demonstrates the use of
//TVirtualStringTree in a multithreading environment.
//The space wiping is done by creating a new file in the root directory of the
//checked drives and filling it with the data from a buffer that has been initialized
//with random data. The filling is done until the whole space of the drive is consumed.
//Once this occurs, the file is deleted. This results in the wiping of all data that
//was left from previous installations.
//The software is licensed under the terms of the
//MIT License
//
//Copyright (c) 2020 Vlker Handel und Dienstleistung UG (haftungsbeschrnkt)
//
//Permission is hereby granted, free of charge, to any person obtaining a copy
//of this software and associated documentation files (the "Software"), to deal
//in the Software without restriction, including without limitation the rights
//to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//copies of the Software, and to permit persons to whom the Software is
//furnished to do so, subject to the following conditions:
//
//The above copyright notice and this permission notice shall be included in all
//copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
//SOFTWARE.


interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls,Generics.Collections,
  Vcl.ExtCtrls, Vcl.CheckLst, VirtualTrees,SyncObjs, Vcl.Menus;

type

  TVolumeInfo = record
    GUIDName:String;
    PathName:String;
    Name:String;
    FileSystemName:String;
    BytesTotal:UInt64;
    BytesFree:UInt64;
    BytesPerBlock:Cardinal;
    State:Byte;
    ProcessingStart:TDateTime;
    Processed:UInt64;
  end;

  STORAGE_ACCESS_ALIGNMENT_DESCRIPTOR = packed record
    Version:DWORD;
    Size:DWORD;
    BytesPerCacheLine:DWORD;
    BytesOffsetForCacheAlignment:DWORD;
    BytesPerLogicalSector:DWORD;
    BytesPerPhysicalSector:DWORD;
    BytesOffsetForSectorAlignment:DWORD;
  end;

const
//STORAGE_PROPERTY_ID enumeration (winioctl.h)
  StorageDeviceProperty                 :DWORD=00;
  StorageAdapterProperty                :DWORD=01;
  StorageDeviceIdProperty               :DWORD=02;
  StorageDeviceUniqueIdProperty         :DWORD=03;
  StorageDeviceWriteCacheProperty       :DWORD=04;
  StorageMiniportProperty               :DWORD=05;
  StorageAccessAlignmentProperty        :DWORD=06;
  StorageDeviceSeekPenaltyProperty      :DWORD=07;
  StorageDeviceTrimProperty             :DWORD=08;
  StorageDeviceWriteAggregationProperty :DWORD=09;
  StorageDeviceDeviceTelemetryProperty  :DWORD=10;
  StorageDeviceLBProvisioningProperty   :DWORD=11;
  StorageDevicePowerProperty            :DWORD=12;
  StorageDeviceCopyOffloadProperty      :DWORD=13;
  StorageDeviceResiliencyProperty       :DWORD=14;
  StorageDeviceMediumProductType        :DWORD=15;
  StorageAdapterRpmbProperty            :DWORD=16;
  StorageAdapterCryptoProperty          :DWORD=17;
  StorageDeviceIoCapabilityProperty     :DWORD=18;
  StorageAdapterProtocolSpecificProperty:DWORD=19;
  StorageDeviceProtocolSpecificProperty :DWORD=20;
  StorageAdapterTemperatureProperty     :DWORD=21;
  StorageDeviceTemperatureProperty      :DWORD=22;
  StorageAdapterPhysicalTopologyProperty:DWORD=23;
  StorageDevicePhysicalTopologyProperty :DWORD=24;
  StorageDeviceAttributesProperty       :DWORD=25;
  StorageDeviceManagementStatus         :DWORD=26;
  StorageAdapterSerialNumberProperty    :DWORD=27;
  StorageDeviceLocationProperty         :DWORD=28;
  StorageDeviceNumaProperty             :DWORD=29;
  StorageDeviceZonedDeviceProperty      :DWORD=30;
  StorageDeviceUnsafeShutdownCount      :DWORD=31;
  StorageDeviceEnduranceProperty        :DWORD=32;

//STORAGE_QUERY_TYPE enumeration (winioctl.h)

  PropertyStandardQuery  :DWORD=0;
  PropertyExistsQuery    :DWORD=1;
  PropertyMaskQuery      :DWORD=2;
  PropertyQueryMaxDefined:DWORD=3;

type
  STORAGE_PROPERTY_QUERY = packed record
    PropertyId:DWord;
    QueryType:DWord;
    AdditionalParameters:array[0..3]of Byte; //lt. Header nur 1 Byte, allerdings spricht die Doku von "Parameters", also mehreren. Unter 4 Byte geht nix.
  end;

  TNodeData = record
    VolumeArrayIndex:LongInt;
  End;

  TMyThread=Class;

  TTerminationRequested = function:Boolean of object;
  TUpdateStatus = procedure(Node:PVirtualNode;NewState:Byte;Processed:UInt64)of object;
  TExecutionFinishedNotify = procedure(Sender:TMyThread)of Object;

  TMyThread = class(TThread)
  private
    FDrive:String;
    FTErminationRequested:TTerminationRequested;
    FUpdateStatus:TUpdateStatus;
    FExecutionFinishedNotify:TExecutionFinishedNotify;
    FNode:PVirtualNode;
    FClusterSize:Cardinal;
  public
    Constructor Create(Drive:String;ClusterSize:Cardinal;TerminationRequested:TTerminationRequested;updateStatus:TUpdateStatus
                       ;ExecutionFinishedNotify:TExecutionFinishedNotify
                       ;Node:PVirtualNode);
    Destructor Destroy;override;
    procedure Execute;override;
  end;


  TfrmWipeSpace = class(TForm)
    GroupBox1: TGroupBox;
    Panel1: TPanel;
    tree: TVirtualStringTree;
    btnStartCancel: TButton;
    tAppUpdate: TTimer;
    MainMenu1: TMainMenu;
    Program1: TMenuItem;
    Program2: TMenuItem;
    Close1: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure treeGetText(Sender: TBaseVirtualTree; Node: PVirtualNode;
      Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
    procedure treeAfterCellPaint(Sender: TBaseVirtualTree;
      TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex;
      CellRect: TRect);
    procedure btnStartCancelClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure tAppUpdateTimer(Sender: TObject);
    procedure Program2Click(Sender: TObject);
    procedure Close1Click(Sender: TObject);
  private
    { Private-Deklarationen }
    FVolumes:array of TVolumeInfo;
    FCS:TCriticalSection;
    FThreads:TList<TMyThread>;
    FThreadListSync:TMultiReadExclusiveWriteSynchronizer;
    FTerminationSync:TMultiReadExclusiveWriteSynchronizer;
    FTerminationRequested:Boolean;
    FCLosePending:Boolean;

    FCloseRequested:TDateTime;

    function isTerminationRequested:Boolean;
    procedure updateStatus(Node:PVirtualNode;NewState:Byte;Processed:UInt64);
    procedure notifyExecutionFinished(Sender:TMyThread);

    function Describe(vi:TVolumeInfo):String;

    function FormatByteCount(NumBytes:UInt64):String;

    procedure EnumVolumes;
    procedure UpdateVolumeInfo;
  public
    { Public-Deklarationen }
  end;

var
  frmWipeSpace: TfrmWipeSpace;

implementation

{$R *.dfm}

uses uInfo;

Constructor TMyThread.Create(Drive:String;ClusterSize:Cardinal;TerminationRequested:TTerminationRequested;updateStatus:TUpdateStatus
                             ;ExecutionFinishedNotify:TExecutionFinishedNotify
                             ;Node:PVirtualNode);
Begin
  inherited create(True);
  FreeOnTerminate:=True;
  FDrive:=Drive;
  FTerminationRequested:=TerminationRequested;
  FUpdateStatus:=updateStatus;
  FexecutionFinishedNotify:=executionFinishedNotify;
  FNode:=Node;
  FCLusterSize:=ClusterSize;
End;


Destructor TMyThread.Destroy;
Begin
  inherited;
End;

procedure TMyThread.Execute;
const cCheckPeriod:Extended=1/(24*60*60*2); //0,5 sek
var sFileName:String;
    buff:TByteArray;
    pFileName:PChar;
    i:LongInt;

    pWord:^Word;

    dtLastCheck:TDateTime;

    bFinish:Boolean;

    iCopySize:Cardinal;

    Buffer:Pointer;
    BufferSize:Cardinal;
    FileStream:TFileStream;

Begin
  Buffer:=nil;
  FileStream:=nil;
  bFinish:=False;
  pFileName:=PChar(@buff);
  if getTempFileNameW(PChar(FDrive),nil,0,pFileName)<>0 then Begin
    sFileName:=pFileName;
    FileStream:=TFileStream.Create(sFileName,fmOpenWrite+fmShareExclusive);
    try
      BufferSize:=5*1024*1024;
      GetMem(Buffer,BufferSize);
      Randomize;
      for I := 0 to (5*1024*1024 div 2)-1 do Begin
        pWord:=Pointer(Cardinal(Buffer)+2*i);
        pWord^:=Random(65536);
      End;
      iCopySize:=BufferSize;
      dtLastCheck:=Now;
      repeat
        try
          iCopySize:=FileStream.Write(Buffer^,BufferSize);
          if iCopySize<>BufferSize then Begin
            bFinish:=True;
            FUpdateStatus(FNode,2,0);
          End;
        except
          bFinish:=True;
          FUpdateStatus(FNode,4,0);
        end;
        if not bFinish and((Now-dtLastCheck)>cCheckPeriod) then Begin
          bFinish:=FTerminationRequested;
          if bFinish then FUpdateStatus(FNode,3,0)
          Else FUpdateStatus(FNode,1,FileStream.Position);
        End;
      until bFinish;
    finally
      FreeAndNil(FileStream);
      DeleteFile(sFileName);
      if Assigned(Buffer) then FreeMem(Buffer,BufferSize);
    end;
  End Else FUpdateStatus(FNode,4,0);
  FExecutionFinishedNotify(Self);
End;


procedure TfrmWipeSpace.btnStartCancelClick(Sender: TObject);
var i:LongInt;
    Node:PVirtualNode;
    PNodeData:^TNodeData;
begin
  if btnStartCancel.Caption='Cancel' then Begin
    FTerminationSync.BeginWrite;
    FTerminationRequested:=True;
    FTerminationSync.EndWrite;
  End Else Begin
    if tree.CheckedCount>0 Then Begin
      FTerminationRequested:=False;
      Node:=tree.GetFirstChecked;
      repeat
        PNodeData:=Node.GetData;
        i:=PNodeData^.VolumeArrayIndex;
        FThreads.Add(TMyThread.Create(FVolumes[i].PathName,FVolumes[i].BytesPerBlock,isTerminationRequested,updateStatus
        ,notifyExecutionFinished
        ,Node));
        Node:=tree.GetNextChecked(Node);
      until not Assigned(Node);
      btnStartCancel.Caption:='Cancel';
      tAppUpdate.Enabled:=True;
      for i := 0 to FThreads.Count-1 do FThreads[i].Start;
    End Else MessageDlg('Please check the volumes you want the free space wiped.',mtWarning,[mbOK],0);
  End;
end;

procedure TfrmWipeSpace.Close1Click(Sender: TObject);
begin
  Close;
end;

function TfrmWipeSpace.Describe(vi: TVolumeInfo): String;
begin
  Result:=vi.GUIDName+';'+vi.PathName+';'+vi.Name+';'+vi.FileSystemName+';'+FormatByteCount(vi.BytesTotal)+';'+FormatByteCount(vi.BytesFree);
end;

procedure TfrmWipeSpace.EnumVolumes;
  function PathValid(path:String):Boolean;
  Begin
    Result:=(Pos('\\?\',path)=1)and(Copy(path,Length(path),1)='\')
  End;
const buffLen=32768 div SizeOf(WideChar);
var buffVolumeName:TByteArray;
    hSearchVolume:THandle;
    sVolumeName:String;
    vi:TVolumeInfo;
    iLen:LongInt;
begin
  vi.Name:='';
  vi.PathName:='';
  vi.Name:='';
  vi.FileSystemName:='';
  vi.BytesTotal:=0;
  vi.BytesFree:=0;
  vi.BytesPerBlock:=0;
  vi.State:=0;
  SetLength(FVolumes,0);
  hSearchVolume:=FindFirstVolumeW(@buffVolumeName,buffLen);
  if hSearchVolume<>INVALID_HANDLE_VALUE then try
    sVolumeName:=PChar(@buffVolumeName);
    if PathValid(sVolumeName) then Begin
      vi.GUIDName:=sVolumeName;
      iLen:=length(FVolumes);
      SetLength(FVolumes,iLen+1);
      FVolumes[iLen]:=vi;
    End;
    while FindNextVolumeW(hSearchVolume,@buffVolumeName,buffLen) do Begin
      sVolumeName:=PChar(@buffVolumeName);
      if PathValid(sVolumeName) then Begin
        vi.GUIDName:=sVolumeName;
        iLen:=length(FVolumes);
        SetLength(FVolumes,iLen+1);
        FVolumes[iLen]:=vi;
      End;
    End;
  finally
    FindVolumeClose(hSearchVolume);
  End;
end;

function TfrmWipeSpace.FormatByteCount(NumBytes: UInt64): String;
begin
  if NumBytes>(1024 * 1024 * 1024) then Result:=Format('%.2n GB',[NumBytes / (1024 * 1024 * 1024)])
  Else if Numbytes>(1024*1024) then Result:=Format('%.0n MB',[NumBytes / (1024 * 1024)])
  Else if Numbytes>(1024) then Result:=Format('%.0n kB',[NumBytes / (1024)])
  Else Result:=Format('%.0n B',[Extended(NumBytes)])
end;

procedure TfrmWipeSpace.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
var i:LongInt;
begin
  FThreadListSync.BeginRead;
  try
    if FThreads.Count=0 then CanClose:=True
    Else if Now-FCloseRequested>(5/24*3600) then Begin//5 sec
//Hart beenden. Buffer und FileStream werden nicht freigegeben. Option wre, Buffer im Thread-Konstruktor zu allokieren und FileStream dort zu erzeugen und beides im Destruktor freizugeben.
//Aufgrund der verwendeten Zeitslots kein Problem
      for i:=FThreads.Count-1DownTo 0 do FThreads[i].Free;
      FThreads.Clear;
    End Else Begin
      CanClose:=False;
      if not FCLosePending then Begin
        FTerminationSync.BeginWrite;
        try
          FTerminationRequested:=True;
        finally
          FTerminationSync.EndWrite;
        end;
        FClosePending:=True;
        FCLoseRequested:=Now;
      End;
    End;
  finally
    FThreadListSync.EndRead;
  end;

end;

procedure TfrmWipeSpace.FormCreate(Sender: TObject);
var i:LongInt;
    pNodEData:^TNodeData;
    aNode:PVirtualNode;
    buff:TByteArray;
    sPath:String;
    pFileName:PChar;

begin
  FCS:=TCriticalSection.Create;
  FTerminationSync:=TMultiReadExclusiveWriteSynchronizer.Create;
  FThreadListSync:=TMultiReadExclusiveWriteSynchronizer.Create;
  FThreads:=TList<TMyThread>.Create;
  FTerminationRequested:=False;
  FClosePending:=False;
  SetLength(FVolumes,0);
  EnumVolumes;
  UpdateVolumeInfo;
  tree.Clear;
  tree.NodeDataSize:=SizeOf(TVolumeInfo);
  for i := 0 to Length(FVolumes)-1 do Begin
    if (FVolumes[i].PathName<>'')and(FVolumes[i].BytesFree>0) Then Begin
      sPath:=IncludeTrailingbackslash(FVolumes[i].PathName);
      pFileName:=PCHar(@buff);
      if getTempFileNameW(PChar(sPath),nil,0,pFileName)<>0 then Begin
        DeleteFile(pFileName);
      End Else Continue;
      aNode:=tree.AddChild(nil);
      aNode.CheckType:=ctCheckBox;
      aNode.CheckState:=csUncheckedNormal;
      pNodeData:=aNode.GetData;
      pNodeData.VolumeArrayIndex:=i;
    End;
  End;
end;

procedure TfrmWipeSpace.FormDestroy(Sender: TObject);
begin
  SetLength(FVolumes,0);
  Finalize(FVolumes);
  FreeAndNil(FCS);
  FReeAndNil(FThreads);
  FreeAndNil(FTerminationSync);
  FreeAndNil(FThreadListSync);
end;

function TfrmWipeSpace.isTerminationRequested: Boolean;
begin
  FTerminationSync.BeginRead;
  Result:=FTerminationRequested;
  FTErminationSync.EndRead;
end;

procedure TfrmWipeSpace.notifyExecutionFinished(Sender: TMyThread);
var i:LongInt;
begin
  FThreadListSync.BeginWrite;
  try
    i:=FThreads.IndexOf(Sender);
    FThreads.Delete(i);
  finally
    FThreadListSync.EndWrite;
  end;
end;

procedure TfrmWipeSpace.Program2Click(Sender: TObject);
begin
  frmInfo.ShowModal;
end;

procedure TfrmWipeSpace.tAppUpdateTimer(Sender: TObject);
var doTerminate:Boolean;
begin
  doTerminate:=False;
  if btnStartCancel.Caption='Cancel' then Begin
    FThreadListSync.BeginRead;
    FTerminationSync.BeginRead;
    try
      if FClosePending then Begin
        if (FThreads.Count=0)or(Now-FCloseRequested>(5/24*3600))  then doTerminate:=True
      End Else if FThreads.Count=0 then Begin
        btnStartCancel.Caption:='Wipe Space!';
        tAppUpdate.Enabled:=False;
      End;
    finally
      FThreadListSync.EndRead;
      FTerminationSync.EndRead;
    end;
  End;
  if doTerminate then Close;
end;

procedure TfrmWipeSpace.treeAfterCellPaint(Sender: TBaseVirtualTree;
  TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex;
  CellRect: TRect);
var i:LongInt;
    pNodeData:^TNodeData;
    PaintRect:TRect;
    iTotalSize,iFilled:UInt64;
    iFillWidth:LongInt;
begin
  if Column in [4,6] then Begin
    PNodeData:=Node.GetData;
    i:=PNodeData^.VolumeArrayIndex;
    FCS.Enter;
    try
      case Column of
        4:With TargetCanvas Do Begin
            PaintRect:=CellRect;
            PaintRect.Inflate(-2,-2);
            Brush.Style:=bsSolid;
            Brush.Color:=clWhite;
            Pen.Color:=clBlue;
            Pen.Width:=1;
            Rectangle(PaintRect);
            Brush.Color:=clBlue;
            iTotalSize:=FVolumes[i].BytesTotal;
            iFilled:=iTotalSize-FVolumes[i].BytesFree;
            iFillWidth:=Round(PaintRect.Width * (iFilled / iTotalSize));
            PaintRect.Width:=iFillWidth;
            RectAngle(PaintRect);
          End;
        6:With TargetCanvas Do Begin
          PaintRect:=CellRect;
          PaintRect.Inflate(-2,-2);
          case FVolumes[i].State of
            0:Begin
                Brush.Style:=bsSolid;
                Brush.Color:=clWhite;
                Pen.Color:=clWhite;
                Rectangle(PaintRect);
                Font.Color:=clBlue;
                TextRect(PaintRect,PaintRect.Left,PaintRect.Top,'not started');
              End;
            1:Begin
                Brush.Style:=bsSolid;
                Brush.Color:=clWhite;
                Pen.Color:=clBlue;
                Pen.Width:=1;
                Rectangle(PaintRect);
                Brush.Color:=clBlue;
                iTotalSize:=FVolumes[i].BytesFree;
                iFilled:=FVolumes[i].Processed;
                iFillWidth:=Round(PAintRect.Width * (iFilled / iTotalSize));
                PaintRect.Width:=iFillWidth;
                RectAngle(PaintRect);
              End;
            2:Begin
                Brush.Style:=bsSolid;
                Brush.Color:=clWhite;
                Pen.Color:=clWhite;
                Rectangle(PaintRect);
                Font.Color:=clGreen;
                TextRect(PaintRect,PaintRect.Left,PaintRect.Top,'Finished');
              End;
            3:Begin
                Brush.Style:=bsSolid;
                Brush.Color:=clWhite;
                Pen.Color:=clWhite;
                Rectangle(PaintRect);
                Font.Color:=clFuchsia;
                TextRect(PaintRect,PaintRect.Left,PaintRect.Top,'Canceled');
              End;
            4:Begin
                Brush.Style:=bsSolid;
                Brush.Color:=clWhite;
                Pen.Color:=clWhite;
                Rectangle(PaintRect);
                Font.Color:=clRed;
                TextRect(PaintRect,PaintRect.Left,PaintRect.Top,'Error!');
              End;
          end;
        End;
      end;
    finally
      FCS.Leave;
    end;
  End;
end;

procedure TfrmWipeSpace.treeGetText(Sender: TBaseVirtualTree; Node: PVirtualNode;
  Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
var pNodeData:^TNodeData;
    i:LonGInt;

begin
  pNodeData:=Node.GetData;
  i:=pNodeData^.VolumeArrayIndex;
  FCS.Enter;
  try
    case Column of
      0:CellText:=FVolumes[i].PathName;                                 //MountPoint
      1:CellText:=FVolumes[i].Name;                                     //Name
      2:CellText:=FVolumes[i].FileSystemName;                           //FS
      3:CellText:=FormatByteCount(FVolumes[i].BytesTotal);              //Size
      4:CellText:='';
      5:CellText:=FormatByteCount(FVolumes[i].BytesPerBlock);           //Block Size
      6:CellText:='';
    end;
  finally
    FCS.Leave;
  end;
end;

procedure TfrmWipeSpace.updateStatus(Node: PVirtualNode; NewState: Byte;
  Processed: UInt64);
var i:LongInt;
    PNodeData:^TNodeData;
begin
  FCS.Enter;
  try
    PNodeData:=Node.GetData;
    i:=PNodeData^.VolumeArrayIndex;
    FVolumes[i].State:=NewState;
    case NewState of
      0:FVolumes[i].Processed:=0;
      1:FVolumes[i].Processed:=Processed;
      2:FVolumes[i].Processed:=FVolumes[i].BytesFree;
      3:FVolumes[i].Processed:=Processed;
    end;
  finally
    FCS.Leave;
  end;
  tree.Invalidate;
end;

procedure TfrmWipeSpace.UpdateVolumeInfo;
const buffLen=32768 div SizeOf(WideChar);
var vi:TVolumeInfo;
    i:LongInt;
    buffMP:TByteArray;
    buffVN:TByteArray;
    buffFSN:TByteArray;
    dwVolumeSerialNumber
    ,dwMaximumComponentlength
    ,dwFileSystemFlags:Cardinal;
    h:THandle;
    iReturnLen:Cardinal;
    iBytesCaller:UInt64;
    saad:STORAGE_ACCESS_ALIGNMENT_DESCRIPTOR;
    spq:STORAGE_PROPERTY_QUERY;
    pSAAD:^STORAGE_ACCESS_ALIGNMENT_DESCRIPTOR;
    pSPQ:^STORAGE_PROPERTY_QUERY;
    iBytesReturned:Cardinal;
    sGuidName:String;
begin
  if Length(FVolumes)>0 then for i := 0 to Length(FVolumes)-1 do BEgin
    vi:=FVolumes[i];
    if GetVolumePathNamesForVolumeNameW(PChar(vi.GUIDName),@buffMP,buffLen,iReturnLen) then vi.PathName:=PChar(@buffMP); //nur ersten. Result ist eig. ein Array von Nullterminierten strings
    if GetVolumeInformationW(PChar(vi.GUIDName),@buffVN,buffLen,@dwVolumeSerialNumber,dwMaximumComponentlength,dwFileSystemFlags,@buffFSN,buffLen) then Begin
      vi.FileSystemName:=PChar(@buffFSN);
      vi.Name:=PChar(@buffVN);
    End;
    if GetDiskFreeSpaceExW(PChar(vi.GUIDName),iBytesCaller,vi.BytesTotal,@(vi.BytesFree)) then ;
    sGuidName:=Copy(vi.GUIDName,1,Length(vi.GUIDName)-1);
    h:=CreateFileW(PChar(sGUIDName),0,FILE_SHARE_READ+FILE_SHARE_WRITE,nil,OPEN_EXISTING,0,0);
    if h<>INVALID_HANDLE_VALUE then try
      spq.PropertyId:=StorageAccessAlignmentProperty;
      spq.QueryType:=PropertyStandardQuery;
      spq.AdditionalParameters[0]:=0;

      pSAAD:=@saad;
      pSPQ:=@SPQ;

      if DeviceIoControl(h,IOCTL_STORAGE_QUERY_PROPERTY
                        ,pSPQ,SizeOf(spq)   //in
                        ,pSAAD,SizeOf(saad),iBytesReturned //out
                        ,nil //nix overlapped
                        ) then vi.BytesPerBlock:=saad.BytesPerLogicalSector

    finally
      CloseHandle(h);
    end;
    FVolumes[i]:=vi;
  End;
end;

end.
